/*
** Apple Macintosh Developer Technical Support
**
** File:		EventLoop.c
** Written by:	Eric Soldan
**
** Copyright  1990-1993 Apple Computer, Inc.
** All rights reserved.
*/

/* You may incorporate this sample code into your applications without
** restriction, though the sample code has been provided "AS IS" and the
** responsibility for its operation is 100% yours.  However, what you are
** not permitted to do is to redistribute the source as "DSC Sample Code"
** after having made changes. If you're going to re-distribute the source,
** we require that you make it clear in the source that the code was
** descended from Apple Sample Code, but that you've made changes. */



/*****************************************************************************/



#include "App.h"			/* Get the application includes/typedefs, etc.	*/
#include "App.protos.h"		/* Get the prototypes for application.			*/



/*****************************************************************************/



extern RgnHandle	gCursorRgn;				/* DTS.Lib..framework global. */
	/* The current cursor region.  The initial cursor region is an empty region,
	** which will cause WaitNextEvent to generate a mouse-moved event, which will
	** cause us to set the cursor for the first time. */

extern Boolean		gQuitApplication;		/* DTS.Lib..framework global. */
	/* This is set to false by Initialize.  When the user selects Quit
	** (and does not abort the quit), then this boolean is set true. */

Boolean		gLowOnMem;
static void	ManageLowMem(void);

#define kRamReserve 0x10000L
#define kRamSlop    0x08000L



/*****************************************************************************/
/*****************************************************************************/

#ifdef applec
#pragma segment DTSDrawSeg1
#endif

/*****************************************************************************/
/*****************************************************************************/



/* Get events forever, and handle them by calling DoEvent.  Get the events by
** calling WaitNextEvent. */

void	EventLoop(void)
{
	EventRecord		event;

	while (!gQuitApplication) {
		ManageLowMem();
		if (!WaitNextEvent(everyEvent, &event, 15, gCursorRgn))
			event.what = nullEvent;
		DoEvent(&event);
	};
}



/*****************************************************************************/



/* This is called to determine if there is enough ram available for various
** application operations.  Typically, if there is not enough, certain menu
** items are dimmed, thus preventing further use of memory. */

static void	ManageLowMem(void)
{
	static Handle	reserveMemHndl;

	if (!reserveMemHndl) {
		reserveMemHndl = NewHandle(0);		/* Create the initial reserve-mem handle. */
		EmptyHandle(reserveMemHndl);
	}

	if (gLowOnMem) {		/* If we were previously flagged as low on ram... */
		ReallocateHandle(reserveMemHndl, kRamReserve + kRamSlop);	/* See if we have enough now.	*/
		if (*reserveMemHndl) {								/* We now have enough.				*/
			SetHandleSize(reserveMemHndl, kRamReserve);		/* Free up the slop.				*/
			HPurge(reserveMemHndl);							/* Make reserveMem purgeable again. */
			gLowOnMem = false;								/* Flag that we have plenty of ram. */
		}
	}
	else {
		if (!*reserveMemHndl) {			/* Somebody found it necessary to purge memory... */
			ReallocateHandle(reserveMemHndl, kRamReserve);		/* Can get get our reserve back? */
			if (*reserveMemHndl) HPurge(reserveMemHndl);		/* Yes, so flag it purgeable. */
			else {
				NewDocumentWindow(nil, 'LMEM', false);			/* Not enough ram.  Warn user.    */
				gLowOnMem = true;								/* Set flag so application knows. */
			}
		}
	}
}



/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/



/*  Called by DTS.Lib..framework.  */

/* Handle the cursor changes.  Most of the work is done by the application
** framework.  Each window has its own cursor calculation proc, so you
** shouldn't have to add any code here.  You may wish to do some special
** stuff prior to the DoWindowCursor call.  That's why this is called first. */

void	DoCursor(void)
{
	DoWindowCursor();		/* Call the framework to figure out the cursor. */
}



